/*
 * arch/mips/dec/int-handler.S
 *
 * Copyright (C) 1995, 1996 Paul M. Antoine
 *
 * Written by Ralf Baechle and Andreas Busse, modified for DECStation
 * support by Paul Antoine.
 *
 * NOTE: There are references to R4X00 code in here, because there is an
 *	 upgrade module for Personal DECStations with such a CPU!
 *
 * FIXME: still plenty to do in this file, as much of the code hasn't been
 *	  modified to suit the DECStation's interrupts.
 */
#include <asm/asm.h>
#include <asm/regdef.h>
#include <asm/fpregdef.h>
#include <asm/mipsconfig.h>
#include <asm/mipsregs.h>
#include <asm/stackframe.h>
#include <asm/bootinfo.h>

/* FIXME: what about other decstations? */
#include <asm/dec/kn01.h>
#include <asm/dec/kn02.h>
#include <asm/dec/kn02ba.h>
#include <asm/dec/kn02ca.h>

#include <asm/dec/interrupts.h>


		.text
		.set	noreorder
/*
 * decstation_handle_int: Interrupt handler for Personal DECStation 5000/2x
 *
 * FIXME: this is *extremely* experimental, though it is probably o.k. for
 *	  most DECStation models.
 *
 * We follow the model in the Indy interrupt code by David Miller, where he
 * says: a lot of complication here is taken away because:
 *
 * 1) We handle one interrupt and return, sitting in a loop
 *    and moving across all the pending IRQ bits in the cause
 *    register is _NOT_ the answer, the common case is one
 *    pending IRQ so optimize in that direction.
 *
 * 2) We need not check against bits in the status register
 *    IRQ mask, that would make this routine slow as hell.
 *
 * 3) Linux only thinks in terms of all IRQs on or all IRQs
 *    off, nothing in between like BSD spl() brain-damage.
 *
 * Furthermore, the IRQs on the DECStations look basically (barring
 * software IRQs which we don't use at all) like...
 *
 * DS2100/3100's, aka kn01, aka Pmax:
 *
 *	MIPS IRQ	Source
 *      --------        ------
 *             0	Software (ignored)
 *             1        Software (ignored)
 *             2        SCSI
 *             3        Lance Ethernet
 *             4        DZ11 serial
 *             5        RTC
 *             6        Memory Controller
 *             7        FPU
 *
 * DS5000/200, aka kn02, aka 3max:
 *
 *	MIPS IRQ	Source
 *      --------        ------
 *             0	Software (ignored)
 *             1        Software (ignored)
 *             2        TurboChannel
 *             3        RTC
 *             4        Reserved
 *             5        Memory Controller
 *             6        Reserved
 *             7        FPU
 *
 * DS5000/1xx's, aka kn02ba, aka 3min:
 *
 *	MIPS IRQ	Source
 *      --------        ------
 *             0	Software (ignored)
 *             1        Software (ignored)
 *             2        TurboChannel Slot 3, system board
 *             3        TurboChannel Slot 2
 *             4        TurboChannel Slot 1
 *             5        TurboChannel Slot 0
 *             6        Halt button
 *             7        FPU
 *
 * DS5000/2x's, aka kn02ca, aka maxine:
 *
 *	MIPS IRQ	Source
 *      --------        ------
 *             0	Software (ignored)
 *             1        Software (ignored)
 *             2        Periodic Interrupt (100usec)
 *             3        RTC
 *             4        I/O write timeout
 *             5        TurboChannel
 *             6        Halt Keycode from Access.Bus keyboard (CTRL-ALT-ENTER)
 *             7        FPU
 *
 * FIXME: We'd like handle the IRQ according to _our_ priority but can't for now.
 * Priority might be:
 *
 * Highest ----     RTC
 *                  SCSI (if separate from TC)
 *                  Ethernet (if separate from TC)
 *                  TurboChannel (if there is one!)
 *		    Serial (if separate from TC)
 *                  Memory Controller
 * Lowest  ----     Halt (if there is one!)
 *
 * then we just return, if multiple IRQs are pending then we will just take
 * another exception, big deal.
 */
		NESTED(decstation_handle_int, PT_SIZE, ra)
		.set	noat
		SAVE_ALL
		move	k0, sp
		CLI
		.set	at

		/*
		 * Get pending interrupts
		 */
		mfc0	t0,CP0_CAUSE		# get pending interrupts
#ifdef 0
		mfc0	t1,CP0_STATUS		# get enabled interrupts
		nop
		and	t0,t1			# isolate allowed ones
		andi	t0,0xff00		# isolate pending bits
/*
 * FIXME: The following branch was:
 *					beqz	t0,spurious_interrupt
 *
 *	  ...but the wonders of ecoff cause the gas assembler (ver 2.5.1 )
 *	  to complain:
 *
 *		  "Can not represent relocation in this object file format"...
 *
 *	  hence this hack to branch foward a bit, and then jump <sigh>
 *	  Perhaps a later version of gas will cope? - Paul
 */
		beqz	t0,3f;
#endif
		sll	t0,16			# delay slot

		/*
		 * Find irq with highest priority
		 * FIXME: This is slow
		 */
		la	t1,ll_vectors
1:		bltz	t0,2f			# found pending irq
		sll	t0,1
		b	1b
		subu	t1,PTRSIZE		# delay slot

		/*
		 * Do the low-level stuff
		 */
2:		lw	t0,(t1)
                nop
		jr	t0
		nop				# delay slot
		END(decstation_handle_int)

/*
 * FIXME: The hack mentioned above.
 */
3:		PANIC("Spurious Interrupt!")
		j	spurious_interrupt
		nop

/*
 * FIXME: Many of these are unimplemented!
 *
 * Interrupt routines common to all DECStations first.
 */
		EXPORT(dec_intr_sw0)
dec_intr_sw0:	PANIC("Unimplemented sw0 handler")

		EXPORT(dec_intr_sw1)
dec_intr_sw1:	PANIC("Unimplemented sw1 handler")

		EXPORT(dec_intr_fp0)
dec_intr_fpu:	PANIC("Unimplemented FPU interrupt handler")

		EXPORT(dec_intr_rtc)
dec_intr_rtc:	li      s1,~IE_IRQ3 
		lui     s3,%hi(intr_count)
		lw      t2,%lo(intr_count)(s3)
		addiu   t2,1
		sw      t2,%lo(intr_count)(s3)
#ifdef 0
		mfc0    t1,CP0_STATUS           # reenable interrupts 
		li      t0,ST0_CU0|1            # without rtc
		and     t1,s1
		or      t1,t0
		mtc0    t1,CP0_STATUS
#endif
		move	a2, k0			# needs pointer to regs on stack!
		jal     timer_interrupt
		nop
#ifdef 0
		mfc0    t1,CP0_STATUS           # disable interrupts again
		li      t0,ST0_CU0|1
		or      t1,t0
		xori    t1,1
		mtc0    t1,CP0_STATUS
		li      t0,IE_IRQ3		# set rtc as allowed interrupt
		or      t1,t0
		mtc0    t1,CP0_STATUS
#endif
		lw	t2,%lo(intr_count)(s3)
		subu	t2,1
#ifdef 0
		j	v0			# entry.S loads this!?!
		sw      t2,%lo(intr_count)(s3)
#endif
		j	ret_from_sys_call
		sw	t2,%lo(intr_count)(s3)  # delay slot

/*
 * Generic unimplemented interrupt routine - vector_table is initialised to
 * point all interrupts here.  The table is then filled in by machine-specific
 * initialisation in dec_setup().
 */
		EXPORT(dec_intr_unimplemented)
dec_intr_unimplemented:
		mfc0	a1,CP0_CAUSE		# cheats way of printing an arg!
		nop				# to be sure...
		PANIC("Unimplemented interrupt! Cause register was 0x%x\n");

/*
 * PMAX interrupt routine prototypes for DS2100/3100's and possible 5100's
 */
		EXPORT(kn01_sii)
kn01_sii:	PANIC("Unimplemented sii handler")
		EXPORT(kn01_lance)
kn01_lance:	PANIC("Unimplemented lance handler")
		EXPORT(kn01_memctrlr)
kn01_memctrlr:	PANIC("Unimplemented memory controller handler")
		EXPORT(kn01_dz)
kn01_dz:	PANIC("Unimplemented dz handler")

/*
 * 3MAX interrupt routine prototypes for DS5000/2xx's
 */
		EXPORT(kn02_memctrlr)
kn02_memctrlr:	PANIC("Unimplemented memory controller interrupt handler")
		EXPORT(kn02_tc)
kn02_tc:	PANIC("Unimplemented TurboChannel interrupt handler")

/*
 * 3MIN interrupt routine prototypes for DS5000/1xx's
 */
		EXPORT(kn02ba_tc3)
kn02ba_tc3:	PANIC("Unimplemented TurboChannel slot 3 interrupt handler")
		EXPORT(kn02ba_tc2)
kn02ba_tc2:	PANIC("Unimplemented TurboChannel slot 2 interrupt handler")
		EXPORT(kn02ba_tc1)
kn02ba_tc1:	PANIC("Unimplemented TurboChannel slot 1 interrupt handler")
		EXPORT(kn02ba_tc0)
kn02ba_tc0:	PANIC("Unimplemented TurboChannel slot 0 interrupt handler")

/*
 * Halt interrupt, initiated by a button on the DS5000/1xx
 */
		EXPORT(kn02ba_halt)
kn02ba_halt:	li      a0,0
		jal     pmax_halt
		li      a1,0                    # delay slot

/*
 * MAXINE interrupt routines for DS5000/2x's
 */
		EXPORT(kn02ca_periodic)
kn02ca_periodic:
		PANIC("Unimplemented periodic interrupt handler")
		EXPORT(kn02ca_tc)
kn02ca_tc:	PANIC("Unimplemented TurboChannel interrupt handler")

/*
 * Halt interrupt, initiated by CTRL-ALT-ENTER on the keyboard of the DS5000/2x
 */
		EXPORT(kn02ca_halt)
kn02ca_halt:
		li      a0,0
		jal     rex_halt
		li      a1,0                    # delay slot

/*
 * Just for debugging... load a0 with address of the point inside the
 * framebuffer at which you want to draw a line of 16x32 pixels.
 * Maxine's framebuffer starts at 0xaa000000.
 */
		.set	reorder
		LEAF(drawline)
		li	t1,0xffffffff		# set all pixels on
		li	t2,0x10			# we will write 16 words
1:		sw	t1,(a0)			# write the first word
		addiu	a0,a0,4			# move our framebuffer pointer
		addiu	t2,t2,-1		# one less to do
		bnez	t2,1b			# finished?
		jr 	ra
		END(drawline)

/*
 *  This routine is need for the memory detection routine dec_get_memory_size.
 *  The routine causes as dbe exception if it read a non existing physical
 *  adress. The exception handler does the rest. 
 */

        LEAF(check_memory)
        beq     a1,zero,1f
        nop
        sw      zero,0(a0)
1:      lb      a1,0(a0)
        nop
        j       ra
        nop
        END(check_memory)

/*
 * FIXME: This interrupt vector table is experimental.  It is initialised with 
 *	  dec_intr_unimplemented and filled in with the addresses of
 *	  machine-specific interrupt routines in dec_setup()  Paul 10/5/97.
 */
		.data
		EXPORT(vector_table)
vector_table:
		PTR	dec_intr_unimplemented
		PTR	dec_intr_unimplemented
		PTR	dec_intr_unimplemented
		PTR	dec_intr_unimplemented
		PTR	dec_intr_unimplemented
		PTR	dec_intr_unimplemented
		PTR	dec_intr_unimplemented
ll_vectors:	PTR	dec_intr_fpu		# FPU, same on all DECStations!

#ifdef 0
/*
 * FIXME: This table needs to be filled in in dec_setup, as it is machine-
 *	  specific!! Paul 4/5/97.
 */
		.data
		PTR	ll_sw0			# SW0
		PTR	ll_sw1			# SW1
		PTR	ll_timer		# 100us Periodic interrupt
		PTR	ll_rtc			# RTC periodic interrupt
		PTR	ll_io_error		# Timeout on I/O writes
		PTR	ll_tc3			# TC slot 3, motherboard
		PTR	ll_reset		# Halt keycode (CTRL+ALT+ENTER)
ll_vectors:	PTR	ll_fpu			# FPU
got_here:	.asciiz	"Got an interrupt!\n";
got_here1:	.asciiz	"Returning from low-level handler!\n";
got_here2:	.asciiz	"Entering timer_int!\n";
got_here3:	.asciiz	"Returning from timer_int!\n";

/* thats for the DS2100/DS3100 Frieder */
		.data
		PTR	ll_sw0			# SW0
		PTR	ll_sw1			# SW1
                PTR     ll_sii                  # SCSI interrupt
                PTR     ll_lance                # Ethernet interrupt
                PTR     ll_dz                   # Serial interrupt,kb,mouse etc.
		PTR	ll_rtc			# RTC periodic interrupt
		PTR	ll_io_error		# Timeout on I/O writes
ll_vectors:	PTR	ll_fpu			# FPU
#endif





