/*
 * drivers/dec/char/dc.h
 *
 * Copyright (C) 1997 Michael Engel
 *
 * Originally written by Michael Engel.
 *
 * The DEC DZ11 serial chip is designated a DC7085 SLU.  This is the header
 * file for the routines that access that chip.
 *
 * FIXME: This was ripped out of decstation.c, and is untested since then.
 */

#ifndef DC_LINES	/* first all-caps define */
/* Serial line registers */
struct dc_reg {
	unsigned short  DZ_CSR;		/* SLU control status register	*/
	unsigned short  pad1[3];
	union {				/* SLU read buffer/line param.	*/
		unsigned short  DZ_RBUF;
		unsigned short  DZ_LPR;
	} DZ_RBUF_LPR;
	unsigned short  pad2[3];
	unsigned short  DZ_TCR;		/* SLU transmitter control reg. */
	unsigned short  pad3[3];
	union {				/* SLU modem status/txmt reg.	*/
		unsigned short 	DZ_MSR;
		unsigned short  DZ_TDR;
	} DZ_MSR_TDR;
};

#define dccsr		sc_regs->DZ_CSR
#define dcdtr		sc_regs->DZ_TCR
#define dcmsr		sc_regs->DZ_MSR_TDR.DZ_MSR
#define dcbrk		sc_regs->DZ_MSR_TDR.DZ_MSR
#define dcrbuf		sc_regs->DZ_RBUF_LPR.DZ_RBUF
#define dclpr		sc_regs->DZ_RBUF_LPR.DZ_LPR
#define dctcr		sc_regs->DZ_TCR
#define dctbuf		sc_regs->DZ_MSR_TDR.DZ_TDR

#define DC_LINES	4		/* four serial lines on the chip */
#define DC_KB		0		/* keyboard line */
#define DC_MS		1		/* mouse line */
#define DC_MODEM	2		/* modem line */
#define DC_PRINTER	3		/* printer line */

/* Interrupt controller register bits */
#define SINT_SR 0200			/* Serial line recv/silo full	*/
#define SINT_ST 0100			/* Serial line transmitter done */

/* Control status register definitions (dccsr) */
#define DC_OFF		0x00		/* Modem control off		*/
#define DC_MAINT	0x08		/* Maintenance			*/
#define DC_CLR		0x10		/* Reset dc7085 chip		*/
#define DC_MSE		0x20		/* Master Scan Enable		*/
#define DC_RIE		0x40		/* Receive IE */
#define DC_RDONE	0x80		/* Receiver done		*/
#define DC_TIE		0x4000		/* Trasmit IE */
#define DC_TRDY		0x8000		/* Transmit ready		*/

/* Line parameter register definitions (dclpr) */
#define BITS5		0x00		/* 5 bit char width		*/
#define BITS6		0x08		/* 6 bit char width		*/
#define BITS7		0x10		/* 7 bit char width		*/
#define BITS8		0x18		/* 8 bit char width		*/
#define TWOSB		0x20		/* two stop bits		*/
#define PENABLE		0x40		/* parity enable		*/
#define OPAR		0x80		/* odd parity			*/
#define DC_B4800	0xc00		/* 4800 BPS speed		*/
#define DC_B9600	0xe00		/* 9600 BPS speed		*/
#define DC_RE		0x1000		/* Receive enable		*/

/* Receiver buffer register definitions (dcrbuf) */
#define DC_PE		0x1000		/* Parity error			*/
#define DC_FE		0x2000		/* Framing error		*/
#define DC_DO		0x4000		/* Data overrun error		*/
#define DC_DVAL		0x8000		/* Receive buffer data valid	*/

/* Line control status definitions (dclcs) */
#define DC_SR		0x08		/* Secondary Receive		*/
#define DC_CTS		0x10		/* Clear To Send		*/
#define DC_CD		0x20		/* Carrier Detect		*/
#define DC_RI		0x40		/* Ring Indicate		*/
#define DC_DSR		0x80		/* Data Set Ready		*/
#define DC_LE		0x100		/* Line Enable			*/
#define DC_DTR		0x200		/* Data Terminal Ready		*/
#define DC_BRK		0x400		/* Break			*/
#define DC_ST		0x800		/* Secondary Transmit		*/
#define DC_RTS		0x1000		/* Request To Send		*/

/* DM lsr definitions */
#define SML_LE		0x01		/* Line enable			*/
#define SML_DTR		0x02		/* Data terminal ready		*/
#define SML_RTS		0x04		/* Request to send		*/
#define SML_ST		0x08		/* Secondary transmit		*/
#define SML_SR		0x10		/* Secondary receive		*/
#define SML_CTS		0x20		/* Clear to send		*/
#define SML_CAR		0x40		/* Carrier detect		*/
#define SML_RNG		0x80		/* Ring				*/
#define SML_DSR		0x100		/* Data set ready, not DM bit	*/

/* dcdtr bits */
#define DC_RDTR 0x400			/* REAL data terminal ready	*/

/* dcmsr bits */
#define DC_RDSR 0x200			/* REAL data set ready bit	*/
#define DC_XMIT DC_RDSR			/* Ready to transmit & rec.	*/

/* #defines to activate the chips transmitters, receivers, etc */

#define dc_master_scan_enable()	(dccsr |= DC_MSE)
#define dc_master_scan_disable() (dccsr &= ~DC_MSE)

#define dc_enable_tx_interrupt() (dccsr |= DC_TIE)
#define dc_disable_tx_interrupt() (dccsr &= ~DC_TIE)

#define dc_enable_rx_interrupt() (dccsr |= DC_RIE)
#define dc_disable_rx_interrupt() (dccsr &= ~DC_RIE)

#define dc_enable_tx(x) (dctcr |= (1 << (x)))
#define dc_disable_tx(x) (dctcr &= ~ (1 << (x)))

#define dc_enable_rx(x) dc_attributes[(x)] |= DC_RE; \
			   dclpr = dc_attributes[(x)]

#define dc_disable_rx(x) dc_attributes[(x)] &= ~ DC_RE; \
			   dclpr = dc_attributes[(x)]

#define dc_set_attributes(x, val) dc_attributes[(x)] = (val) + (x); \
			   dclpr = dc_attributes[(x)]

#define dc_get_attributes(x) (dc_attributes[(x)])

#define dc_get_tx_line(x) (((x) & 0x0300) >8)

#define dc_get_rx_line(x) (((x) & 0x0300) >8)

/* synopsis of routines in dc.c */
extern int dc_default_tx_handler();
extern int dc_default_rx_handler();
extern init_dc();
extern dc_handler();
extern unsigned set_dc_tx_handler();
extern unsigned set_dc_rx_handler();
/*extern struct dc_reg *sc_regs;*/
unsigned short dc_attributes[DC_LINES];

/* Monochrome definitions */
#define SELF_TEST	'T'
#define INCREMENTAL	'R'
#define PROMPT		'D'

#define MOUSE_ID	0x2
#define TABLET_ID	0x4

#define START_FRAME	0x80		/* start of report frame bit	*/
#define X_SIGN		0x10		/* sign bit for X		*/
#define Y_SIGN		0x08		/* sign bit for Y		*/

#define UPDATE_POS	0x01

/* Interrupt controller register bits */
#define SINT_VF		010

/* Cursor Command Register bits */
#define ENPA	0000001
#define FOPA	0000002
#define ENPB	0000004
#define FOPB	0000010
#define XHAIR	0000020
#define XHCLP	0000040
#define XHCL1	0000100
#define XHWID	0000200
#define ENRG1	0000400
#define FORG1	0001000
#define ENRG2	0002000
#define FORG2	0004000
#define LODSA	0010000
#define VBHI	0020000
#define HSHI	0040000
#define TEST	0100000


/* Line Prameter Register bits */
#define SER_KBD      000000
#define SER_POINTER  000001
#define SER_COMLINE  000002
#define SER_PRINTER  000003
#define SER_CHARW    000030
#define SER_STOP     000040
#define SER_PARENB   000100
#define SER_ODDPAR   000200
#define SER_SPEED    006000
#define SER_RXENAB   010000

#endif
